var flatpickr = function (selector, config) {
    'use strict';
    var elements,
        createInstance,
        instances = [],
        i;

    createInstance = function (element) {
        if (element._flatpickr) {
            element._flatpickr.destroy();
        }
        element._flatpickr = new flatpickr_(element, config);
        return element._flatpickr;
    };

    if (selector.nodeName)
        return createInstance(selector);
    /*
    Utilize the performance of native getters if applicable
    https://jsperf.com/getelementsbyclassname-vs-queryselectorall/18
    https://jsperf.com/jquery-vs-javascript-performance-comparison/22
    */
    else if ( /^\#[a-zA-Z0-9\-\_]*$/.test(selector) )
        return createInstance(document.getElementById(selector.slice(1)));

    else if ( /^\.[a-zA-Z0-9\-\_]*$/.test(selector) )
        elements = document.getElementsByClassName(selector.slice(1));    
    
    else
        elements = document.querySelectorAll(selector);
    

    for (i = 0; i < elements.length; i++)
        instances.push(createInstance(elements[i]));    

    return {
        calendars: instances,
        byID: function(id){
            for (let calendar of this.calendars) {
                if (calendar.element.id === id)
                    return calendar;
            }
        }
    };
};

class flatpickr_ {

    constructor(element, instanceConfig){

        this.config = {};
        this.l10n = flatpickr.l10n;
        instanceConfig=instanceConfig||{};
        this.element = element;

        for (let config in flatpickr.defaultConfig)
            this.config[config] =
                instanceConfig[config] ||
                this.element.dataset&&this.element.dataset[config.toLowerCase()] ||
                flatpickr.defaultConfig[config];

        this.calendarContainer = document.createElement('div');
        this.calendarContainer.className = 'flatpickr-calendar';
        this.navigationCurrentMonth = document.createElement('span');
        this.navigationCurrentMonth.className = 'flatpickr-current-month';
        this.monthsNav = document.createElement('div');
        this.prevMonthNav = document.createElement('span');
        this.nextMonthNav = document.createElement('span');
        this.calendar = document.createElement('table');
        this.calendarBody = document.createElement('tbody');
        this.currentDate = new Date();
        this.wrapperElement = document.createElement('div');
        this.input = (this.config.wrap) ? element.querySelector("[data-input]") : element; 

        if(this.config.defaultDate){
            this.config.defaultDate = this.uDate(this.config.defaultDate);
        }

        if ( this.input.value || this.config.defaultDate )
            this.selectedDateObj = this.uDate(this.config.defaultDate||this.input.value);

        this.config.minDate =
            this.uDate(this.config.minDate === "today" ? new Date() : this.config.minDate, true);

        this.config.maxDate = this.uDate(this.config.maxDate, true);

        this.jumpToDate(this.selectedDateObj||this.config.minDate||this.currentDate);

        this.wrap();        
        this.buildCalendar();
        this.bind();

        if (!this.config.noCalendar)
            this.updateValue();

    }



    uDate (date, timeless){

        timeless = timeless||false;

        // dashes to slashes because firefox sucks
        if (typeof date === 'string')
            date = new Date(date.replace(/(\d*)[\-\.](\d*)[\-\.](\d*)/g, "$1/$2/$3"));

        timeless && date && ( date.setHours(0,0,0,0) );

        return date;
    }

    equalDates (date1, date2){
        return date1.getFullYear() === date2.getFullYear() &&
                date1.getMonth() === date2.getMonth() &&
                date1.getDate() === date2.getDate();

    }

    wrap  () {

        this.wrapperElement.className = 'flatpickr-wrapper';

        this.element.parentNode.insertBefore(this.wrapperElement, this.element);
        this.wrapperElement.appendChild(this.element);

        if(this.config.inline)
            this.wrapperElement.classList.add('inline');

        if ( this.config.altInput ){
            // replicate this.element
            this.altInput = document.createElement(this.input.nodeName);
            this.altInput.placeholder = this.input.placeholder;
            this.altInput.type = this.input.type||"text";
            
            this.input.type='hidden';
            this.wrapperElement.appendChild(this.altInput);
        }

    };

    getDaysinMonth (givenMonth){

        let yr = this.currentYear, month = givenMonth || this.currentMonth;

        if (month === 1 && ( !( (yr % 4) || (!(yr % 100) && (yr % 400))) ) )
            return 29;

        return this.l10n.daysInMonth[month];
    }

    updateValue (){

        if (this.selectedDateObj && this.config.enableTime ){

            // update time
            var hour = parseInt(this.hourElement.value),
                minute = (60+parseInt(this.minuteElement.value ))%60;

            if (!this.config.time_24hr)
                hour = hour%12 + 12*(this.am_pm.innerHTML=== "PM");

            this.selectedDateObj.setHours(hour , minute );

            this.hourElement.value = 
                this.pad( this.config.time_24hr ? hour : ((12 + hour)%12+12*(hour%12===0)));

            this.minuteElement.value = this.pad(minute);

        }                

        if ( this.altInput && this.selectedDateObj )
            this.altInput.value = this.formatDate(this.config.altFormat);
        
        if ( this.selectedDateObj )
            this.input.value = this.formatDate(this.config.dateFormat);

        this.triggerChange();

    }

    pad(num){ return ("0" + num).slice(-2)}

    formatDate (dateFormat) {

        if (this.config.noCalendar)
            dateFormat = "";

        if(this.config.enableTime)
            dateFormat+= " " + this.config.timeFormat;

        let formattedDate = '',
            formats = {                
                D: () => this.l10n.weekdays.shorthand[ formats.w() ],
                F: () => monthToStr( formats.n() - 1, false ),
                H: () => this.pad(this.selectedDateObj.getHours()),
                K: () => this.selectedDateObj.getHours() > 11 ? "PM" : "AM",
                M: () => monthToStr( formats.n() - 1, true ),
                U: () => this.selectedDateObj.getTime() / 1000,
                Y: () => this.selectedDateObj.getFullYear(),
                d: () => this.pad( formats.j() ),
                h: () => this.selectedDateObj.getHours()%12 ? this.selectedDateObj.getHours()%12 : 12,                
                i: () => this.pad( this.selectedDateObj.getMinutes() ),
                j: () => this.selectedDateObj.getDate(),
                l: () => this.l10n.weekdays.longhand[ formats.w() ],
                m: () => this.pad( formats.n() ),                
                n: () => this.selectedDateObj.getMonth() + 1,
                w: () => this.selectedDateObj.getDay(),
                y: () => String( formats.Y() ).substring(2)          
            },
            formatPieces = dateFormat.split('');

        for(let i = 0; i < formatPieces.length; i++){

            if (formats[formatPieces[i]] && formatPieces[i - 1] !== '\\') 
                formattedDate += formats[formatPieces[i]]();

            else if (formatPieces[i] !== '\\') 
                    formattedDate += formatPieces[i];
        }

        return formattedDate;
    };

    monthToStr (date, shorthand) {

        return shorthand ? this.l10n.months.shorthand[date] : this.l10n.months.longhand[date];

    };


    isDisabled (date){

        for (var i = 0; i < this.config.disable.length; i++)
            if ( date >= this.uDate( this.config.disable[i]['from'] )  && date <= this.uDate( this.config.disable[i]['to'] ) )
                    return true;

        return false;

    };

    


    timeWrapper (e){

        e.preventDefault();

        var min = parseInt(e.target.min),
            max = parseInt(e.target.max),
            step = parseInt(e.target.step),
            delta = step * ( Math.max(-1, Math.min(1, (e.wheelDelta || -e.detail))) ),
            newValue = ( parseInt(e.target.value) + delta )%(max+(min===0));

        newValue < min && ( newValue = max+(min===0) - step*(min===0) );
        e.target.value = this.pad( newValue );

    }


    updateNavigationCurrentMonth () {

        this.navigationCurrentMonth.innerHTML =
            '<span>' +
                this.monthToStr(this.currentMonth, this.config.shorthandCurrentMonth) +
            '</span> '
            + this.currentYear;

    };

    handleYearChange () {

        if (this.currentMonth < 0 || this.currentMonth > 11) {

            this.currentYear += this.currentMonth % 11;
            this.currentMonth = (this.currentMonth + 12) % 12;

        }

    };

    documentClick (event) {

        if (this.wrapperElement && this.wrapperElement.classList.contains("open") && !this.wrapperElement.contains(event.target))
            this.close();

    };

    changeMonth(offset) {
        this.currentMonth += offset;

        this.handleYearChange();
        this.updateNavigationCurrentMonth();
        this.buildDays();
    }

    calendarClick(event) {

        event.preventDefault();

        var targetDate = event.target;

        if ( targetDate.classList.contains('slot') || targetDate.parentNode.classList.contains('slot') )
        {

            this.selectedDateObj = new Date( 
                this.currentYear, this.currentMonth,
                targetDate.childNodes[0].innerHTML||targetDate.innerHTML
            );

            updateValue();
            buildDays();

            if ( !this.config.inline && !this.config.enableTime )
                this.close();            
            
        }

    };

    buildCalendar() {

        if ( !this.config.noCalendar) {
            this.buildMonthNavigation();
            this.buildWeekdays();
            this.buildDays();
            this.calendar.appendChild(this.calendarBody);
            this.calendarContainer.appendChild(this.calendar);
        }            

        this.wrapperElement.appendChild(this.calendarContainer);
        this.config.enableTime && ( this.buildTime() );

    };

    buildMonthNavigation () {

        this.monthsNav.className = 'flatpickr-months';

        this.prevMonthNav.className = "flatpickr-prev-month";
        this.prevMonthNav.innerHTML = this.config.prevArrow;

        this.nextMonthNav.className = "flatpickr-next-month";
        this.nextMonthNav.innerHTML = this.config.nextArrow;

        this.monthsNav.appendChild(this.prevMonthNav);
        this.monthsNav.appendChild(this.navigationCurrentMonth);
        this.monthsNav.appendChild(this.nextMonthNav);

        this.updateNavigationCurrentMonth();
        this.calendarContainer.appendChild(this.monthsNav);
    };

    buildWeekdays() {

        let weekdayContainer = document.createElement('thead'),
            weekdayRow = document.createElement("tr"),
            firstDayOfWeek = this.l10n.firstDayOfWeek,
            weekdays = this.l10n.weekdays.shorthand.slice();

        if (firstDayOfWeek > 0 && firstDayOfWeek < weekdays.length) {
            weekdays = [].concat(weekdays.splice(firstDayOfWeek, weekdays.length), weekdays.splice(0, firstDayOfWeek));
        }
        for(let i = 0; i < weekdays.length; i++){
            let weekday = document.createElement("th");
            weekday.innerText=weekdays[i];
            weekdayRow.appendChild(weekday);
        //weekdayContainer.innerHTML = '<tr><th>' + weekdays.join('</th><th>') + '</th></tr>';        
        }
        weekdayContainer.appendChild(weekdayRow);
        this.calendar.appendChild(weekdayContainer);
    };

    buildDays () {

        let firstOfMonth = ( new Date(this.currentYear, this.currentMonth, 1).getDay() - this.l10n.firstDayOfWeek + 7 )%7,
            numDays = this.getDaysinMonth(),
            prevMonth = ( this.currentMonth - 1 + 12)%12,
            prevMonthDays = this.getDaysinMonth( prevMonth ),
            calendarFragment = document.createDocumentFragment(),
            row = document.createElement('tr'),
            dayNumber = prevMonthDays + 1 - firstOfMonth,
            className,
            cur_date,
            date_is_disabled,
            date_outside_minmax;

        // prepend days from the ending of previous month
        for( ; dayNumber <= prevMonthDays; dayNumber++ ){
             let cell = document.createElement("td"),
                 cell_c = document.createElement("span");
            cell.className = "disabled";
            cell_c.className = "flatpickr-day";
            cell_c.innerText = dayNumber;
            cell.appendChild(cell_c);
            row.appendChild(cell);
        }

        this.calendarBody.innerHTML = '';

        // Start at 1 since there is no 0th day
        for (dayNumber = 1; dayNumber <= 42 - firstOfMonth; dayNumber++) {

            if (dayNumber <= numDays) // avoids new date objects for appended dates
                cur_date = new Date(this.currentYear, this.currentMonth, dayNumber);

            // we have reached the end of a week, wrap to the next line
            if ( (dayNumber + firstOfMonth - 1) % 7 === 0) {
                calendarFragment.appendChild(row);
                row = document.createElement('tr');
            }

            date_outside_minmax = (this.config.minDate && cur_date < this.config.minDate )
                ||  (this.config.maxDate && cur_date > this.config.maxDate);


            date_is_disabled = dayNumber > numDays || date_outside_minmax || this.isDisabled( cur_date );

            className = date_is_disabled ? "disabled" : "slot";

            if (!date_is_disabled && !this.selectedDateObj && this.equalDates(cur_date, this.currentDate) )
                className += ' today';

            if (!date_is_disabled && this.selectedDateObj && this.equalDates(cur_date, this.selectedDateObj) )
                className += ' selected';

            let cell = document.createElement("td"),
                cell_c = document.createElement("span");
            cell.className = className;
            cell_c.className = "flatpickr-day";
            cell_c.innerText = (dayNumber > numDays ? dayNumber % numDays : dayNumber);
            cell.appendChild(cell_c);
            row.appendChild(cell);


        }

        calendarFragment.appendChild(row);
        this.calendarBody.appendChild(calendarFragment);

    };

    buildTime(){

        let timeContainer = document.createElement("div"),
            separator = document.createElement("span");

        timeContainer.className = "flatpickr-time";

        this.hourElement = document.createElement("input");
        this.minuteElement = document.createElement("input");
        

        separator.className = "flatpickr-time-separator";
        separator.innerHTML = ":";

        this.hourElement.className = "flatpickr-hour";
        this.hourElement.type = this.minuteElement.type = "number";
        this.minuteElement.className = "flatpickr-minute";


        this.hourElement.value = 
            this.selectedDateObj ? this.pad(this.selectedDateObj.getHours()) : 12;
        this.minuteElement.value = 
            this.selectedDateObj ? this.pad(this.selectedDateObj.getMinutes()) : "00";


        this.hourElement.step = this.config.hourIncrement;
        this.minuteElement.step = this.config.minuteIncrement;

        this.hourElement.min = +!this.config.time_24hr; // 0 in 24hr mode, 1 in 12hr mode
        this.hourElement.max = this.config.time_24hr ? 23 : 12;

        this.minuteElement.min = 0;
        this.minuteElement.max = 59;               

        this.hourElement.title = this.minuteElement.title = "Scroll to increment";

        timeContainer.appendChild(this.hourElement);
        timeContainer.appendChild(separator);
        timeContainer.appendChild(this.minuteElement);

        if (!this.config.time_24hr){ // add am_pm if appropriate
            this.am_pm = document.createElement("span");
            this.am_pm.className = "flatpickr-am-pm";
            this.am_pm.innerHTML = ["AM","PM"][(this.hourElement.value > 11)|0];
            this.am_pm.title="Click to toggle";
            timeContainer.appendChild(this.am_pm);
        }

        // picking time only
        if (this.config.noCalendar)
            this.selectedDateObj = new Date();        

        this.calendarContainer.appendChild(timeContainer);

    }

    bind () {
        let self = this._flatpickr;
        function am_pm_toggle(e){ 
            e.preventDefault();
            self.am_pm.innerHTML =  ["AM","PM"][(am_pm.innerHTML === "AM")|0]; 
        }

        if (String(self.config.clickOpens)==='true'){
            self.input.addEventListener( 'focus' , self.open);

            if(self.altInput)
                self.altInput.addEventListener( 'focus' , self.open);
        }

        if (self.config.wrap && self.element.querySelector("[data-open]"))
            self.element.querySelector("[data-open]").addEventListener( 'click' , self.open);

        if (self.config.wrap && self.element.querySelector("[data-close]"))
            self.element.querySelector("[data-close]").addEventListener( 'click' , self.close);

        if (self.config.wrap && self.element.querySelector("[data-toggle]"))
            self.element.querySelector("[data-toggle]").addEventListener( 'click' , self.toggle);

        if (self.config.wrap && self.element.querySelector("[data-clear]"))
            self.element.querySelector("[data-clear]").addEventListener( 'click' , self.clear);
        

        self.prevMonthNav.addEventListener('click', ()=>self.changeMonth(-1) );
        self.nextMonthNav.addEventListener('click', ()=> self.changeMonth(1) );

        self.calendar.addEventListener('click', self.calendarClick);
        document.addEventListener('click', self.documentClick);

        if ( self.config.enableTime ){

            self.hourElement.addEventListener("mousewheel", self.timeWrapper);
            self.hourElement.addEventListener("DOMMouseScroll", self.timeWrapper);
            self.minuteElement.addEventListener("mousewheel", self.timeWrapper);
            self.minuteElement.addEventListener("DOMMouseScroll", self.timeWrapper);

            self.hourElement.addEventListener("mouseout", self.updateValue);
            self.minuteElement.addEventListener("mouseout", self.updateValue);

            self.hourElement.addEventListener("change", self.updateValue);
            self.minuteElement.addEventListener("change", self.updateValue);

            self.hourElement.addEventListener("click", () => self.hourElement.select());
            self.minuteElement.addEventListener("click", () => self.minuteElement.select());

            if (!self.config.time_24hr) {
                self.am_pm.addEventListener("focus", () =>  self.am_pm.blur());
                self.am_pm.addEventListener("click", self.am_pm_toggle);

                self.am_pm.addEventListener("mousewheel", self.am_pm_toggle);
                self.am_pm.addEventListener("DOMMouseScroll", self.am_pm_toggle);
                self.am_pm.addEventListener("mouseout", self.updateValue);
            }            

        }
        if(document.createEvent){
            self.clickEvt = document.createEvent("MouseEvent");
            self.clickEvt.initMouseEvent("click",true,true,window,0,0,0,0,0,false,false,false,false,0,null);
        }
        else
            self.clickEvt = new MouseEvent('click', {
                'view': window,
                'bubbles': true,
                'cancelable': true
            });
        

    };

    open() {

        if (this._flatpickr.input.disabled || this._flatpickr.config.inline)
            return;

        this._flatpickr.input.blur();
        this._flatpickr.input.classList.add('active');

        if(this._flatpickr.altInput){
            this._flatpickr.altInput.blur();
            this._flatpickr.altInput.classList.add('active');
        }

        this._flatpickr.element.parentNode.classList.add('open');

    };

    toggle () {
        if (this.input.disabled)
            return;

        this.element.parentNode.classList.toggle('open');

        if(this.altInput)
            this.altInput.classList.toggle('active');

        this.input.classList.toggle('active');
    };

    close () {
        this.element.parentNode.classList.remove('open');
        this.input.classList.remove('active');
        if (this.altInput)
            this.altInput.classList.remove('active');
        this.config.onClose();
    };

    clear() {
        this.input.value="";
        this.selectedDateObj = null;
        this.jumpToDate();
        this.redraw();
    }

    triggerChange(){

        if (this.clickEvt) 
            this.input.dispatchEvent(this.clickEvt)
        else // IE
            this.input.fireEvent("onchange");        

        this.config.onChange(this.selectedDateObj, this.input.value);

    }

    destroy  () {
        let parent = this.element.parentNode,
            element  = parent.removeChild(this.element);

        document.removeEventListener('click', documentClick, false);

        parent.removeChild(calendarContainer);
        parent.parentNode.replaceChild(element, parent);
    };

    jumpToDate (jumpDate) {
        jumpDate = this.uDate(
            jumpDate||
            this.selectedDateObj||this.config.defaultDate||this.config.minDate||
            currentDate
        );
        this.currentYear = jumpDate.getFullYear();
        this.currentMonth = jumpDate.getMonth();
        this.redraw();

    };

    redraw(){

        this.updateNavigationCurrentMonth();
        this.buildDays();

    }



    setDate(date, triggerChangeEvent){

        this.selectedDateObj = uDate(date);
        this.jumpToDate(this.selectedDateObj);
        this.updateValue();

        triggerChangeEvent = triggerChangeEvent||false;
        triggerChangeEvent && (this.triggerChange());

     }


}

flatpickr.l10n = {
    weekdays: {
        shorthand: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
        longhand: ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday']
    },
    months: {
        shorthand: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
        longhand: ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']
    },
    daysInMonth: [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
    firstDayOfWeek: 0
};

flatpickr.defaultConfig = {
    noCalendar: false,
    wrap: false,
    clickOpens: true,
    dateFormat: 'Y-m-d',
    altInput: false,
    altFormat: "F j, Y",
    defaultDate: null,
    minDate: null,
    maxDate: null,
    disable: [],
    shorthandCurrentMonth: false,
    inline: false,
    prevArrow: '&lt;',
    nextArrow: '&gt;',
    enableTime: false,            
    timeFormat: "h:i K",
    time_24hr: false,
    hourIncrement: 1,
    minuteIncrement: 5,
    onChange: function( dateObj, dateStr ){},
    onClose: function() {}
}

